"""
Plot files with gnuplot & PDFLaTeX
- Previously matplotlib was used to plot
- Requires the following packages
  - gnuplot
  - pdflatex
  - cairo
- On MacOS these can all be installed with brew
"""


import subprocess
from string import Template
import argparse
import yaml
import pandas as pd
import sys


def readHeader(filename,header_location):
  """
  Inputs:
  - str  filename: Filename of history file
  Outputs:
  - list header:   list of variables used in dataset, this is used for retrieving indexes 
    when plotting
  """
  file = open(filename)
  for i, line in enumerate(file):
    if i == header_location:
      raw_header = line.strip()[1:]
      header = raw_header.split(",")
      for n in range(len(header)):
        header[n] = header[n].strip()
      break
  return header 

def maskData(data):
  """
  - Occasionally, the simulation may crash due to a Riemann solver failure
  - This is typically fixed by a restart from a checkpoint with a lower courant number
  - However, the data is still there in the history file
  - This function masks out incorrect data, by finding areas where "time travel" (time
    value going back on itself) has occurred
  - Algorithm could be more optimal/use pandas datatypes, but is fine for most cases
  - Looped until data is monotonic, so can handle multiple crash-restart events
  """
  while data["time"].is_monotonic_increasing == False:
    # Run through data to find section where data is not behaving monotonically
    cum_max         = 0 # Cumulative maximum time
    i               = 0 # Index counter
    tl_finish       = 0 # Time when normal timeline interrupted 
    tl_finish_index = 0 # Index when normal timeline interrupted
    for t in data["time"]:
      if t < cum_max:
        if tl_finish_index == 0:
          tl_finish_index = i
          tl_finish       = t
      else:
        if tl_finish_index != 0:
          break
        cum_max = t
      i += 1
    # Run through data again to find start of first time loop
    i = 0
    for t in data["time"]:
      if t > tl_finish:
        tl_start_index = i
        break
      i += 1
    # Delete first  recurrence, so that data works correctly
    data = data.drop(data.index[tl_start_index:tl_finish_index])
  # After data removed, sort to ensure consistency
  # data = data.sort_values(by="time")
  return data

def skipToLastEntry(filename):
  """
  - One quirk of Athena++ history files is that if a new simulation is run with the same
    history filename, data is appended to the end, rather than overwriting the file. 
  - Because I frequently forget to remove all the data from a directory, this data 
    remains, rather than fixing this, I'll just write a function to remove the data from 
    plotting.
  - This function finds all instances where a new athena file has been appended, and
    redurns the row indexes of the header files and start of data for that instance.
  - This is performed by finding the initialisation string of the history file by going
    through the entire dataset, and updating the location indices when a new instance is
    found.
  Potential modification, include in config file a way of changing which data entry to use
  Though right now I cannot think of a reason to do this

  Inputs:
  - str filename: Filename of history file
  Outputs:
  - int header_location: Location of the header row for most recent athena iteration
  - int data_location:   Location of the first data row for most recent athena iteration
  """
  fp = open(filename)
  for i, line in enumerate(fp):
    if line.strip() == "# Athena++ history data":
      header_location = i+1
      data_location   = i+2
  return header_location,data_location

def readData(filename):
  """
  Inputs:
  - str filename: Filename of history file
  Outputs:
  - df data:   Entire dataset from 
  """
  # First, find number of rows to skip
  header_location,data_location = skipToLastEntry(filename)
  # Read in header, then data, then merge into a single pandas array
  header       = readHeader(filename,header_location)
  data         = pd.read_csv(filename,comment="#",skiprows=data_location)
  data.columns = header
  # Check to see if data contains any restarts, mask out data between restarts
  while data["time"].is_monotonic_increasing == False:
    data = maskData(data)
  return data

def evalQuantity(plot,sim_data):
  """
  Evaluate the quantity of plot data
  This plotting program allows for arbitrary mathematical operations in the config
  file, these are 
  Multiple equations are allowed for each plot, allowing for easy comparative
  operations, each evaluation must be separated with a semicolon (";")
  """
  # Split quantities up by semicolon
  evals  = plot["quantity"].split(";") 
  if "qlabels" in plot:
    plot_labels = plot["qlabels"].split(";")
    for label in plot_labels:
      # Clean up labels
      label = label.strip()
  else:
    plot_labels = None
  plot_data = []
  for ev in evals:
    eval_data  = sim_data.eval(ev)
    y_scale    = float(defineParameter(plot,"y_scale",fallback=1.0))
    eval_data *= y_scale
    plot_data.append(eval_data)
  # Quick check to make sure that all labels are correctly applied
  # Labels can be left blank, but if labels are included, need one for each
  if plot_labels != None:
    if len(plot_data) != len(plot_labels):
      print("!!! Mismatch in number of labels and number of comparative plots")
      sys.exit()
  return plot_data

def defineParameter(dictionary, key, fallback = None):
  """
  Define a variable if it is contained in the dictionary, if not use a fallback
  Fairly self explanitory, this gets used a *lot*
  """
  if key in dictionary:
    parameter = dictionary[key]
  else:
    parameter = fallback
  return parameter

def makePlot(gnuplot_file,tex_file):
  subprocess.check_output(["gnuplot",gnuplot_file])
  subprocess.check_output(["pdflatex",tex_file])
  return

def cleanUp(gnuplot_file,tex_file,data_file):
  """
  Remove datafiles and clean up
  """
  return

def readConfigFile(config_filename):
  with open(config_filename) as config_file:
    config = yaml.load(config_file,Loader=yaml.FullLoader)
  return config

def writeGnuPlotHeader(gnuplot_file,
                       template_filename,
                       plot_config):
  """
  Write gnuplot contents that aren't the plot commands, such as setting xlabels and such.
  """
  with open(template_filename) as template_file:
    template_text = template_file.read()
  
  # Conditional template settings, ones which require setting or unsetting manually
  # Logarithmic plotting
  if plot_config["log"] == True:
    plot_config["log_y"] = "set logscale y"
  else:
    plot_config["log_y"] = "unset logscale y"
  # Blank units
  x_unit = defineParameter(plot_config,"x_unit")
  y_unit = defineParameter(plot_config,"y_unit")
  if x_unit == None:
    plot_config["x_unit_proc"] = ""
  else:
    plot_config["x_unit_proc"] = "({})".format(x_unit)
  if y_unit == None:
    plot_config["y_unit_proc"] = ""
  else:
    plot_config["y_unit_proc"] = "({})".format(y_unit)
  # x and y ranges
  x_min = defineParameter(plot_config,"x_min")
  x_max = defineParameter(plot_config,"x_max")
  y_min = defineParameter(plot_config,"y_min")
  y_max = defineParameter(plot_config,"y_max")
  if x_min == None or x_max == None:
    plot_config["x_range"] = "# Do not set x range"
  else:
    plot_config["x_range"] = "set xrange [{0}:{1}]".format(x_min,x_max)
  if y_min == None or y_max == None:
    plot_config["y_range"] = "# Do not set y range"
  else:
    plot_config["y_range"] = "set yrange [{0}:{1}]".format(y_min,y_max)
    
  
  template     = Template(template_text)
  template_sub = template.substitute(plot_config)
  gnuplot_file.write(template_sub)
  
def main(**kwargs):
  config_filename = kwargs["config_filename"]
  config = readConfigFile(config_filename)

  global_config = config["global"]
  history_files = config["history_files"]

  template_filename = defineParameter(global_config,"template",fallback="template.pytmp")

  data = []
  history_filenames = []

  x_quantity = defineParameter(global_config,"x_quantity",fallback="time")
  x_scale    = float(defineParameter(global_config,"x_scale",fallback=1.0))

  for history_file in history_files:
    d = readData(history_file["filename"])
    d = d.sort_values(by=x_quantity)
    data.append(d)
    history_filenames.append(history_file["filename"])


  gnuplot_filenames  = []
  pdflatex_filenames = []

  for plot in config["plots"]:
    # Merge plot dictionary and global dictionary
    plot_config = {**plot,**global_config}
    y_quantity = defineParameter(plot_config,"quantity")
    y_scale    = float(defineParameter(plot_config,"y_scale",fallback=1.0))
    # Generate general filename
    gnuplot_prefix = defineParameter(plot_config,"prefix",fallback="gnuplot")
    gnuplot_suffix = defineParameter(plot_config,"filename",fallback=y_quantity)
    # Generate filename for gnuplot file
    gnuplot_filename = "{}-{}.gp".format(gnuplot_prefix,gnuplot_suffix)
    gnuplot_filenames.append(gnuplot_filename)
    # Generate .tex file filename and add to config
    tex_filename = "{}-{}.tex".format(gnuplot_prefix,gnuplot_suffix)
    pdflatex_filenames.append(tex_filename)
    plot_config["tex_filename"] = tex_filename
    # Generate .csv file filename
    csv_filename = "{}-{}.csv".format(gnuplot_prefix,gnuplot_suffix)
    # Write data files
    plot_data = pd.DataFrame()
    labels = []
    for n in range(len(data)):
      history_file = history_files[n]
      history_filename = defineParameter(history_file,"filename")
      label = defineParameter(history_file,"label",fallback=history_filename)
      sim_xdata  = data[n][x_quantity] * x_scale
      sim_ydata = evalQuantity(plot_config,data[n])
      sim_ydata  = sim_ydata[0]
      sim_xlabel = label + "_" + x_quantity
      sim_ylabel = label + "_" + gnuplot_suffix
      labels.append(sim_xlabel)
      labels.append(sim_ylabel)
      plot_data = pd.concat([plot_data,sim_xdata],axis=1,ignore_index=True)
      plot_data = pd.concat([plot_data,sim_ydata],axis=1,ignore_index=True)
    plot_data.columns = labels
    plot_data.to_csv(csv_filename,index=False)

    # Write gnuplot file
    with open(gnuplot_filename,"w") as gnuplot_file:
      # Write header
      writeGnuPlotHeader(gnuplot_file,template_filename,plot_config)
      # Write lines for each datafile
      gnuplot_file.write("\nplot")
      for n in range(len(history_files)):
        history_file = history_files[n]
        history_filename = defineParameter(history_file,"filename")
        label = defineParameter(history_file,"label",fallback =history_filename)
        n_ls   = n+1
        n_xdata = (n*2)+1
        n_ydata = (n*2)+2
        plot_template = " '{0}' using {1}:{2} w l ls {3} title '{4}'".format(csv_filename,n_xdata,n_ydata,n_ls,label)
        gnuplot_file.write(plot_template)
        if n_ls != len(history_files):
          gnuplot_file.write(",\\\n    ")

  print("> Finished generating plot files!")
  if kwargs["exec"]:
    print ("> Executing plotting!")
    for n in range(len(gnuplot_filenames)):
      subprocess.check_output(["gnuplot",gnuplot_filenames[n]])
      subprocess.check_output(["pdflatex",pdflatex_filenames[n]])
    print("> Compiled plots!")
  # if kwargs["cleanup"]:
  #   print("> Cleaning up stray pdflatex files using wildcards!")
  #   subprocess.check_output(["rm","*.tex"])
  #   subprocess.check_output(["rm","*.gp"])
  #   subprocess.check_output(["rm","*-inc.pdf"])
  #   subprocess.check_output(["rm","*.csv"])
  return

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="Process 2D and hst data from a folder of Athena outputs.\nRequires a config file and a positional argument of at least one config file, additional prefixes and directories can also be added")
  parser.add_argument("config_filename",
                       help="name of yaml config file to be used.")
  parser.add_argument("-p",
                      "--prefix",
                      type=str,
                      help="Add prefix to filename for plot outputs, by default does not use a prefix, prefix ammended in the form of foo.bar.ext, by default bar.ext where bar is the plot name derived from the config file.")
  parser.add_argument("-d",
                      "--dir",
                      type=str,
                      help="Save to directory, by default uses current working directory.")
  parser.add_argument("-e",
                      "--exec",
                      action="store_true",
                      help="Execute gnuplot and pdflatex to build the plot")
  # parser.add_argument("-c",
  #                     "--cleanup",
  #                     action="store_true",
  #                     help="After building plot, delete files, requires --exec flag")
  args = parser.parse_args()
  main(**vars(args))