import argparse
from math import log
import yaml
from plot_lib import cropData, defineScale, multiPlotCropData, reduceQuant, reduceXYZ, readParameter, readDataFile
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm

  
def calcTemperature(data):
  """
  Special plot case: calculating temperature, requires multiple input grids, so higher memory usage
  - Uses the ideal gas law equation to find temperature based on density, pressure and
    mass of a particle in the gas
  """
  kboltz  = 1.380658e-16
  avgMass = 1.67e-24
  tempGrid = (data["press"] * avgMass) / (data["rho"] * kboltz)
  return tempGrid

def calcDustDensity(data):
  """"
  Special plot case: Calculating dust temperature
  """
  rhoDgrid = data["rho"] * data["r1"]
  return rhoDgrid

def main(**kwargs):
 
  # Read in config file
  with open(kwargs["config_file"]) as file:
    config = yaml.load(file,Loader=yaml.FullLoader)
  # Read in files
  models = []
  for file in config["hdf_files"]:
    filename = readParameter(file,"filename",quit_on_fail=True)
    models.append(readDataFile(filename))
  # Use latex text renderer for nice quality plots
  latex_enabled = readParameter(config,"latex",fallback=False)
  if latex_enabled:
    plt.rcParams.update({
    "text.usetex": True,
    "font.family": "serif"})
  # Extract contiguous and scaled X and Y axis data from arrays
  xyscale = readParameter(config,"xyscale",fallback=1.0)
  cmap = readParameter(config,"cmap",fallback="viridis")
  xw = readParameter(config,"plotwidth",fallback=5)
  xh = readParameter(config,"plotheight",fallback=3)
  cbarasp = readParameter(config,"cbar_aspect",fallback="auto")
  for model in models:
    xydata = reduceXYZ(model,scale=xyscale)
    model["xdata"] = xydata[0]
    model["ydata"] = xydata[1]
  # Get columns and rows
  ncols = readParameter(config,"ncols",fallback=len(models))
  nrows = readParameter(config,"nrows",fallback=1)
  # Loop through all plots to be generated
  for plot in config["plots"]:
    quant   = plot["quantity"]
    qscale  = readParameter(plot,"scale",fallback=1.0)
    fig = plt.figure(figsize=(xw,xh)) ## Initialise plotting environment
    # if len(models) == 1:
    #   ax = [ax]
    # Loop through, slightly jank but so is matplotlib
    col = 0
    row = 0
    axes = []
    for n,model in enumerate(models):
      # Determine if any special cases need to be calculated
      if quant == "temp":
        model["temp"] = calcTemperature(model)
      if quant == "rhod":
        model["rhod"] = calcDustDensity(model)
      # Extract contiguous and scaled quantity data from meshblocks
      qdata = reduceQuant(model,quant,scale=qscale)
      # Set subplot environment
      asp = len(model["xdata"])/len(model["ydata"]) # Calculate aspect ratio for plot, assuming cells are square
      xyunit = readParameter(config,"xyunit",fallback="")
      if xyunit != "":
        xyunit = "("+xyunit+")"
      ax = fig.add_subplot(nrows,ncols,n+1)

      ax.set_xlabel("X {}".format(xyunit))
      ax.set_ylabel("Y {}".format(xyunit))
      plotlabel = r"{}".format(readParameter(config["hdf_files"][n],"label",fallback=""))

      if bool(readParameter(config["hdf_files"][n],"show_phase")) == True:
        time = model["Time"]
        period = float(readParameter(config["hdf_files"][n],"period",fallback=0.0))
        offset = float(readParameter(config["hdf_files"][n],"phase_offset",fallback=0.0))
        phase = time/period
        phase += offset
        # print(time,period,phase)
        if plotlabel != "":
          plotlabel = plotlabel + r", "
        plotlabel = plotlabel+r"$\phi$ = {:.3f}".format(phase)
        # print(plotlabel)

      ax.set_title(plotlabel)
      ax.set(aspect=asp) # Set aspect ratio
      # Set range for plots, effectively cropping it
      xrange,yrange = cropData(model["xdata"],model["ydata"],config)
      # Check individual model for x and y range data, slightly jank
      xrange,yrange = multiPlotCropData(xrange,yrange,config["hdf_files"][n])

      ax.set_xlim(xrange)
      ax.set_ylim(yrange)
      # Render plot using delauney triangulation
      qmin  = readParameter(plot,"qmin")
      qmax  = readParameter(plot,"qmax")
      log   = readParameter(plot,"log",fallback=False)
      cbar  = defineScale(log,qmin,qmax,qdata)

      if (readParameter(config["hdf_files"][n],"xmin") != None):
        ax.set_yticks([])
      ax.label_outer()

      if log:
        qdata = np.clip(qdata,cbar.qmintri,cbar.qmaxtri)
        tri = ax.tricontourf(model["xdata"],model["ydata"],qdata,      
                               levels=cbar.lcont,
                               norm=LogNorm(vmin=cbar.qmintri,vmax=cbar.qmaxtri),
                               cmap=cmap)
      else:
        tri = ax.tricontourf(model["xdata"],model["ydata"],qdata,
                               levels=cbar.lcont,
                               cmap=cmap)
              
      for c in tri.collections:
        c.set_rasterized(True)
      
      # x = []
      # y = []
      # z = []
      # with open("mesh_structure.dat") as f:
      #   for line in f:
      #       if line[0] != '\n' and line[0] != '#':
      #           numbers_str = line.split()
      #           x.append(float(numbers_str[0])*xyscale)
      #           y.append(float(numbers_str[1])*xyscale)
      #           # append zero if 2D
      #           if(len(numbers_str) > 2):
      #               z.append(float(numbers_str[2])*xyscale)
      #           else:
      #               z.append(0.0)
      #       if line[0] == '\n' and len(x) != 0:
      #           ax.plot(x, y, 'w',linewidth=0.5)
      #           x = []
      #           y = []
      #           z = []

      axes.append(ax)
      # Horrible hack bodge to reconcile 1D list of models with MPL column/row list
      if col == ncols-1:
        row += 1
        col = 0
      else:
        col += 1

    
      

    # # Finish up, save plot
    # for a in ax.flat:
    #   a.label_outer()
    label = readParameter(plot,"label",fallback="")
    fig.colorbar(tri, ax=axes,ticks=cbar.lcbar,label=label,aspect=cbarasp)
    dpi = readParameter(config,"dpi",fallback=300)
    fprefix = readParameter(config,"filename",fallback="plot")
    fsuffix = readParameter(plot,"quantity",fallback="a")

    plt.savefig("{}-{}.pdf".format(fprefix,fsuffix),dpi=dpi,bbox_inches="tight")
    print("> Finished {}!".format(quant))
    del fig,ax # Delete plotting environment just in case

  return

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="Process 2D and hst data from a folder of Athena outputs.\nPositional arguments and flags can  be used, if all arguments are left blank, a config file will instead be used.")
  parser.add_argument("config_file",
                       help="name of yaml config file to be used.")
  parser.add_argument("hdf5_files",
                      nargs="*",
                      help="Filenames of hdf5 files, if left blank, program will attempt to find any history files in current folder.")
  args = parser.parse_args()
  main(**vars(args))
