from matplotlib.colors import LogNorm
import athena_read
from numba import njit,prange
import numpy as np
import matplotlib.pyplot as plt
from math import ceil, floor, log10
import sys

def readDataFile(filename):
  """
  Read in an HDF5 datafile, this is just a wrapper
  """
  data = athena_read.athdf(filename,raw = True)
  return data

def multiPlotCropData(xrange,yrange,model):
  # Get x and y min and max plotting values for specific HDF5 file
  xmin = readParameter(model,"xmin",fallback=None)
  xmax = readParameter(model,"xmax",fallback=None)
  ymin = readParameter(model,"ymin",fallback=None)
  ymax = readParameter(model,"ymax",fallback=None)
  # If any values are changed in model plot, use them instead
  if xmin != None:
    xrange[0] = xmin
  if xmax != None:
    xrange[1] = xmax
  if ymin != None:
    yrange[0] = ymin
  if ymax != None:
    yrange[1] = ymax
  # Return x and y ranges 
  return xrange,yrange

def cropData(xdata,ydata,config):
  xmin = readParameter(config,"xmin",fallback=None)
  xmax = readParameter(config,"xmax",fallback=None)
  ymin = readParameter(config,"ymin",fallback=None)
  ymax = readParameter(config,"ymax",fallback=None)
  # If values are not specified, calculate from minimum and maximum
  if xmin == None:
    xmin = np.min(xdata)
  if xmax == None:
    xmax = np.max(xdata)
  if ymin == None:
    ymin = np.min(ydata)
  if ymax == None:
    ymax = np.max(ydata)
  # Build x and y range lists
  xrange = [xmin,xmax]
  yrange = [ymin,ymax]
  # Finish, return x and y range 
  return xrange,yrange

class ColourBar:
  def __init__(self,lcont,lcbar,qmintri,qmaxtri):
    self.lcont = lcont
    self.lcbar = lcbar
    self.qmintri = qmintri
    self.qmaxtri = qmaxtri

def defineScale(log,qmin=None,qmax=None,qdata=None,contourlevels=256):
  """
  Define logarithmic or non-logarithmic scale
  INPUTS:
  - bool log: Is scale logarithmic?
  - 
  """
  # If qmin and qmax are not provided, find values from data
  if qmax == None and qmin == None:
    if type(qdata) == None:
      print("Function defineScale either needs data or qmin+qmax")
      sys.exit()
  if qmin == None:
    qmin = np.min(qdata)
  if qmax == None:
    qmax = np.max(qdata)
  # Calculate spacing
  if log:
    qmin = floor(log10(float(qmin)))
    qmax = ceil(log10(float(qmax)))
    qdel = (qmax - qmin) + 1
    lcont = np.logspace(qmin,qmax,contourlevels)
    lcbar= np.logspace(qmin,qmax,qdel)
    qmintri = 10**qmin
    qmaxtri = 10**qmax
  else:
    lcont = np.linspace(qmin,qmax,contourlevels)
    lcbar = np.linspace(qmin,qmax,6)
    qmintri = qmin
    qmaxtri = qmax
  # Put values into class and return
  cbar = ColourBar(lcont,lcbar,qmintri,qmaxtri)
  return cbar

def plotTriContourf(xdata,ydata,qdata,cbar,log=False,cmap="plasma",raster=True):
  ax = plt.gca()
  if log:
    qdata = np.clip(qdata,cbar.qmintri,cbar.qmaxtri)
    tri = ax.tricontourf(xdata,ydata,qdata,
                          levels=cbar.lcont,
                          norm=LogNorm(vmin=cbar.qmintri,vmax=cbar.qmaxtri),
                          cmap=cmap)
  else:
    tri = ax.tricontourf(xdata,ydata,qdata,
                          levels=cbar.lcont,
                          cmap=cmap)
  # Ensure rasterised
  if raster:
    for c in tri.collections:
      c.set_rasterized(True)
  return tri

def readParameter(dictionary,key,fallback=None,quit_on_fail=False):
  """
  Define a variable if it is contained in the dictionary, if not use a fallback or quit outright.

  Inputs:
    - dict dictionary:   dictionary to read
    - str  key:          key to read
    - val  fallbac:      fallback to return instead, default Nonetype
    - bool quit_on_fail: if True, halt program if key not present,
                         default False
  """
  try:
    parameter = dictionary[key]
  except:
    if quit_on_fail == True:
      print("! Essential variable not found in dictionary, exiting.")
      raise
    else:
      parameter = fallback
      pass
  return parameter

def reduceXYZ(data,scale = 1.0):
  # Embedded function used as this section requires
  # JIT compilation
  @njit(parallel = True)
  def reduce(xgrid,ygrid,nb,ni,nj,nk,nr,scale):
    reduced_grid = np.zeros((3,nr))
    # Split parallelism across coarsest section of loop, this offers a significant speedup
    for n in prange(nb):
      for i in range(ni):
        for j in range(nj):
          for k in range(nk):
            idx = n + (i*nb) + (j*nb*ni) + (k*nb*ni*nj)
            reduced_grid[0,idx] = xgrid[n][i] * scale
            reduced_grid[1,idx] = ygrid[n][j] * scale
            reduced_grid[2,idx] = zgrid[n][k] * scale
    return reduced_grid

  xgrid = data["x1v"]
  ygrid = data["x2v"]
  zgrid = data["x3v"]
  # Determine length of arrays
  nb = len(xgrid) # Total number of meshblocks
  ni = len(xgrid[0]) # X meshblock size
  nj = len(ygrid[0]) # Y meshblock size
  nk = len(zgrid[0]) # Z meshblock size
  # Determine required length of reduced grid
  nr = nb * ni * nj * nk
  reduced_grid = reduce(xgrid,ygrid,nb,ni,nj,nk,nr,scale)
  return reduced_grid

def reduceQuant(data,quantity,scale = 1.0):
  @njit(parallel = True)
  def reduce(qgrid,nn,ni,nj,nk,nr,scale):
    q_reduced = np.zeros(nr)
    for n in prange(nn):
      for i in range(ni):
        for j in range(nj):
          for k in range(nk):
            idx = n + (i*nn) + (j*nn*ni) + (k*nn*ni*nj)
            q_reduced[idx] = qgrid[n][k][j][i] * scale 
    return q_reduced
  qgrid = data[quantity]
  nb = len(data["x1v"])
  ni = len(data["x1v"][0])
  nj = len(data["x2v"][0])
  nk = len(data["x3v"][0])
  nr = nb * ni * nj * nk
  reduced_grid = reduce(qgrid,nb,ni,nj,nk,nr,scale)
  return reduced_grid